/*-
 * Copyright (c) 2008 FUKAUMI Naoki.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/ioctl.h>
#include <err.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>

#include <signal.h>

#include "usb_linux.h"

#define IDVENDOR 0x3275
#define IDPRODUCT 0x6051
#define EP1 (USB_ENDPOINT_IN | 1)
#define BUFFERSIZE 8192


static int interrupted = 0;

static void sig_handler( int sig )
{
    if( sig != SIGALRM && sig != SIGPIPE ){
        fprintf( stderr,  "interrupted\n" );
        interrupted = 1;
    }
    signal( sig, sig_handler );
}


int
main(int argc, char *argv[])
{
    usb_dev_handle* ugen = NULL;

	uint8_t data[0x200], r5 = 0x91, saddr = 0x5a;
	int catv = 0, chan, freq, null = 0, val;
	int xpid_min = 0x1fff, xpid_max = 0x1fff;

	if (argc != 2) {
            fprintf(stderr, "usage: %s channel\n", argv[0] );
		exit(EXIT_FAILURE);
	}

	if (argv[1][0] == 'X') {
		null = 1; /* output XOR'ed null packet only */
		xpid_min = 0x0000;
		xpid_max = 0x1ffe;
		argv[1]++;
	}

	if (argv[1][0] == 'X') {
		null = 2; /* output null packet which contains XOR table */
		argv[1]++;
	}

	if (argv[1][0] == 'C') {
		catv = 1;
		argv[1]++;
	}
	chan = strtoul(argv[1], NULL, 0);

	if (catv == 1) {
		if (chan < 36 || chan > 63) {
			fprintf(stderr, "invalid channel\n");
			goto err;
		}
		freq = 303 + (chan - 36) * 6;
	} else {
		if (chan < 13 || chan > 62) {
			fprintf(stderr, "invalid channel\n");
			goto err;
		}
		freq = 473 + (chan - 13) * 6;
	}

	fprintf(stderr, "channel: %s%d%s\n", (catv == 1) ? "C" : "", chan,
	    (null > 0) ? " (null packet only)" : "");
	fprintf(stderr, "freq: %d MHz\n", freq);

        if( signal( SIGHUP, sig_handler ) == SIG_IGN ) signal( SIGHUP, SIG_IGN );
        if( signal( SIGINT, sig_handler ) == SIG_IGN ) signal( SIGINT, SIG_IGN );

        ugen = init_device( IDVENDOR, IDPRODUCT );
        if( ! ugen ) goto err;

	send_ctl(data, ugen, 0x0c, 0x0000, 0x0000, 0x003a);
	send_ctl(data, ugen, 0x17, 0x0002, 0x0000, 0x0003);
	send_ctl(data, ugen, 0x08, 0xfb07, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x10, 0x0200, 0x0000, 0x0001);

	/* Power off */
	send_ctl(data, ugen, 0x08, 0x0100, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0xdd0c, 0x0000, 0x0001);

	/* send_ctl(data, ugen, 0x0c, 0x0000, 0x0000, 0x003a); */
	/* send_ctl(data, ugen, 0x01, 0x000b, 0x0000, 0x0001); */
	send_ctl(data, ugen, 0x17, 0x0002, 0x0000, 0x0003);

	/* Tuner on */
	send_ctl(data, ugen, 0x08, 0xd500, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x0800, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x4040, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x8080, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x0404, 0x0000, 0x0001);

	/* Tuner init */
	send_ctl(data, ugen, 0x02, 0x00a8, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0120, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0220, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0320, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0420, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0520, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0620, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0720, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0820, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0c20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1120, 0x0021, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1220, 0x000b, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1320, 0x002c, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1420, 0x000a, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1520, 0x0040, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1620, 0x000a, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1720, 0x0070, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1820, 0x0031, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1920, 0x0013, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1a20, 0x0031, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1b20, 0x0013, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1c20, 0x002a, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1d20, 0x00aa, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1e20, 0x00a2, 0x0002);
	send_ctl(data, ugen, 0x03, 0x1f20, 0x0008, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2020, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2120, 0x00ff, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2220, 0x0080, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2320, 0x004c, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2420, 0x004c, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2520, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2620, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2720, 0x000c, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2820, 0x0060, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2920, 0x006b, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2a20, 0x0040, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2b20, 0x0040, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2c20, 0x00ff, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2d20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2e20, 0x00ff, 0x0002);
	send_ctl(data, ugen, 0x03, 0x2f20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3020, 0x0028, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3120, 0x000f, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3220, 0x00a0, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3420, 0x003f, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3820, 0x0001, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3920, 0x0070, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3a20, 0x002d, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3b20, 0x0021, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3c20, 0x003f, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3d20, 0x0010, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3e20, 0x0008, 0x0002);
	send_ctl(data, ugen, 0x03, 0x3f20, 0x000c, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4020, 0x000c, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4120, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4220, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4320, 0x004f, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4420, 0x00ff, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4620, 0x0020, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4720, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4820, 0x0090, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4920, 0x00e6, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4a20, 0x0002, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4b20, 0x0054, 0x0002);
	send_ctl(data, ugen, 0x03, 0x4c20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5020, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5120, 0x0058, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5220, 0x0020, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5420, 0x0057, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5520, 0x00f1, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5620, 0x0020, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5720, 0x0070, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5820, 0x0060, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5c20, 0x0050, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5d20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x5f20, 0x0080, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7020, 0x0018, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7120, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7220, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7520, 0x0022, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7620, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7720, 0x0001, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7c20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x7d20, 0x0052, 0x0002);
	send_ctl(data, ugen, 0x03, 0xba20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0xbb20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0xbc20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0xc220, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0xc720, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0xe420, 0x001a, 0x0002);
	send_ctl(data, ugen, 0x03, 0xe920, 0x0008, 0x0002);
	send_ctl(data, ugen, 0x03, 0xec20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0xef20, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x02, 0x00a8, 0x0000, 0x0002);

	/* Random values */
	send_ctl(data, ugen, 0x1a, 0x0000, 0x0000, 0x0010);

	/* ASIE5606 on */
	send_ctl(data, ugen, 0x08, 0x11ff, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x1100, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x11ff, 0x0000, 0x0001);

	/* send_ctl(data, ugen, 0x02, 0x0400 | saddr, 0x0000, 0x0002); */
	send_ctl(data, ugen, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
	if (data[0] != 1) {
		saddr = 0x4a;
		r5 = 0x80;
		send_ctl(data, ugen, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
	}
	send_ctl(data, ugen, 0x23, 0x0000 | saddr, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x02, 0x0900 | saddr, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x09, 0x0100, 0x0000, 0x0001);

	send_ctl(data, ugen, 0x04, 0x0040, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x05, (data[1] << 8) | 0x0340, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x05, (xpid_min & 0xff00) | 0x41,
	    (xpid_min & 0xff), 0x0003);
	send_ctl(data, ugen, 0x05, (xpid_max & 0xff00) | 0x43,
	    (xpid_max & 0xff), 0x0003);

	send_ctl(data, ugen, 0x0a, 0x0000, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x03, 0x0500 | saddr, r5, 0x0002);

	/* Set frequency */
	val = (freq + 57) * 7;
	send_ctl(data, ugen, 0x0d, 0xfe00, (val & 0xff00) | 0xc0, 0x0004);
	send_ctl(data, ugen, 0x0d, ((val & 0xff) << 8) | 0x03,
	    (freq < 420) ? 0x8291 : 0x8491, 0x0004);
	send_ctl(data, ugen, 0x0d, 0xe306, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x0e, 0x0020, 0x0000, 0x0008);
	send_ctl(data, ugen, 0x03, 0x0120, 0x0040, 0x0002);

	send_ctl(data, ugen, 0x04, 0x0000, 0x0000, 0x0042);
	val = data[65];
	send_ctl(data, ugen, 0x09, 0x0100, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x05, 0x003f, (val & 0xff) | 0x04, 0x0003);
	send_ctl(data, ugen, 0x05, (xpid_min & 0xff00) | 0x41,
	    (xpid_min & 0xff), 0x0003);
	send_ctl(data, ugen, 0x05, (xpid_max & 0xff00) | 0x43,
	    (xpid_max & 0xff), 0x0003);

	send_ctl(data, ugen, 0x06, 0x0000, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x04, 0x0040, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x05, 0x0f40, 0x0000, 0x0002);

	/* Set XOR values which is used against XOR table */
	for (val = 0x1000; val < 0x2000; val += 0x0100)
		send_ctl(data, ugen, 0x03, (val & 0xff00) | saddr,
		    (null == 2) ? 0xff : 0x00, 0x0002);

        /* --------------------------- */
        fprintf( stderr, "start\n" );

        char buffer[ BUFFERSIZE ];
        int i;
        while( ! interrupted ){

            int ret = usb_bulk_read( ugen, EP1, buffer, BUFFERSIZE, 5000 );
            if( ret != BUFFERSIZE ){
                fprintf( stderr, "failed to read by bulk\n" );
                break;
            }
            write( STDOUT_FILENO, buffer, ret );
        }

        fprintf( stderr, "stop\n" );
        /* --------------------------- */

	/* Stop */
	send_ctl(data, ugen, 0x07, 0x0000, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x09, 0x0100, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x03, 0x8000 | saddr, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0400 | saddr, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x03, 0x0000 | saddr, 0x0000, 0x0002);
	send_ctl(data, ugen, 0x21, 0x0000 | saddr, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0x0100, 0x0000, 0x0001);
	send_ctl(data, ugen, 0x08, 0xdd0c, 0x0000, 0x0001);

        close_device( ugen );
	return EXIT_SUCCESS;

 err:

        close_device( ugen );
	return EXIT_FAILURE;
}
